import os
import json
import jsonschema
from unittest import TestCase
import tempfile

import delta.manifest.manifest as manifest
from delta.manifest.manifest import manifest_version, manifest_schema_filename


def _json_sample_path(version='1.0'):
    return os.path.join(os.path.dirname(__file__))


class TestManifest(TestCase):
    schema = None

    @classmethod
    def setUpClass(cls) -> None:
        # Load manifest schema: consider json is correct
        path = os.path.join(manifest._schema_path("1.0"),
                            manifest_schema_filename)
        fp = open(path)
        cls.schema = json.load(fp)
        print(f"Loaded Schema from {path}: {cls.schema}.")
        fp.close()

    def test_read_manifest(self):
        manifest_dict = manifest.read_manifest(os.path.join(
            _json_sample_path(), 'test_manifest_full.json'))
        self.assertEqual(manifest_dict['name'], 'Delta Twin name')

    def test_write_manifest_minimal(self):
        file_path = manifest.write_manifest()
        self.assertTrue(os.path.exists(file_path))
        data = manifest.read_manifest(file_path)
        os.remove(file_path)
        self.assertTrue(manifest.check_manifest(data))

    def test_write_manifest_to_dest(self):
        my_file_destination = tempfile.NamedTemporaryFile().name
        file_path = manifest.write_manifest(path=my_file_destination)
        self.assertEqual(my_file_destination, file_path)
        self.assertTrue(os.path.exists(file_path))
        data = manifest.read_manifest(file_path)
        os.remove(my_file_destination)
        self.assertTrue(manifest.check_manifest(data))

    def test_write_manifest_with_version(self):
        my_file_destination = tempfile.NamedTemporaryFile().name

        file_path = manifest.write_manifest(path=my_file_destination,
                                            version=manifest_version)

        self.assertEqual(my_file_destination, file_path)
        self.assertTrue(os.path.exists(file_path))
        data = manifest.read_manifest(file_path)
        os.remove(my_file_destination)
        self.assertTrue(manifest.check_manifest(data))

    def test_write_manifest_with_bad_version(self):
        my_file_destination = tempfile.NamedTemporaryFile().name

        with self.assertRaises(manifest.ManifestException):
            file_path = manifest.write_manifest(path=my_file_destination,
                                                version="0.0")
        self.assertFalse(os.path.exists(my_file_destination))

    def test_0_check_schema(self):
        # Checks sample manifest with schema
        jsonschema.Draft202012Validator.check_schema(self.schema)

    def test_10_check_manifest_ko(self):
        path = os.path.join(_json_sample_path(), 'test_manifest_ko.json')
        self.assertFalse(manifest.check_manifest(path))

    def test_10_check_manifest_ko_verbose(self):
        path = os.path.join(_json_sample_path(), 'test_manifest_ko.json')
        with open(path, 'r') as f:
            data = json.load(f)
            self.assertFalse(manifest.check_manifest(data, verbose=True))

    def test_10_check_manifest_ok(self):
        path = os.path.join(_json_sample_path(), 'test_manifest_minimum.json')
        with open(path, 'r') as f:
            data = json.load(f)
            self.assertTrue(manifest.check_manifest(data,
                                                    version='1.0'))

    def test_10_check_manifest_ok_verbose(self):
        path = os.path.join(_json_sample_path(), 'test_manifest_minimum.json')
        with open(path, 'r') as f:
            data = json.load(f)
            self.assertTrue(manifest.check_manifest(data,
                                                    verbose=True,
                                                    version='1.0'))

    def test_1_check_sample_full(self):
        # Load manifest sample
        path = os.path.join(_json_sample_path(), 'test_manifest_full.json')
        with open(path) as fp:
            # Checks sample manifest with schema
            jsonschema.validate(instance=json.load(fp), schema=self.schema)

    def test_1_check_sample_minimum(self):
        # Load manifest sample
        path = os.path.join(_json_sample_path(), 'test_manifest_minimum.json')
        with open(path) as fp:
            # Checks sample manifest with schema
            jsonschema.validate(instance=json.load(fp), schema=self.schema)

    def test_2_check_model_retrieval(self):
        # Load manifest sample
        path = os.path.join(_json_sample_path(), 'test_manifest_full.json')
        m = manifest.read_manifest(path)
        self.assertEqual(m['models']['type'], 'nifi')
        self.assertEqual(m['models']['parameters']['parameters1'], 10)
        self.assertEqual(m['models']['parameters']['parameters2'], 'hello')
        self.assertEqual(m['models']['parameters']['parameters3']['Dict']
                         ['subparameters1'], 10)



"""database creation

Revision ID: 825d896d998a
Revises: 
Create Date: 2024-01-22 15:27:04.701809

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '825d896d998a'
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('runs',
    sa.Column('id', sa.String(), nullable=False),
    sa.Column('deltatwin_id', sa.String(), nullable=False),
    sa.Column('deltatwin_version', sa.String(), nullable=False),
    sa.Column('owner', sa.String(), nullable=False),
    sa.Column('date_created', sa.DateTime(), nullable=False),
    sa.Column('status', sa.Enum('CREATED', 'RUNNING', 'SUCCEEDED', 'CANCELLED', 'ERROR', name='runstatus'), nullable=False),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('parameters',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('run_id', sa.String(), nullable=False),
    sa.Column('name', sa.String(), nullable=False),
    sa.Column('param_type', sa.String(), nullable=False),
    sa.Column('type', sa.String(), nullable=False),
    sa.Column('kind', sa.Enum('INPUT', 'OUTPUT', name='parameterkind'), nullable=False),
    sa.Column('value', sa.String(), nullable=True),
    sa.Column('url', sa.String(), nullable=True),
    sa.Column('path', sa.String(), nullable=True),
    sa.Column('size', sa.Integer(), nullable=True),
    sa.Column('checksum', sa.String(), nullable=True),
    sa.ForeignKeyConstraint(['run_id'], ['runs.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('parameters')
    op.drop_table('runs')
    # ### end Alembic commands ###



from delta.manifest.parser import (
    Dependency,
    Input,
    License,
    Manifest,
    ManifestParser,
    Model,
    Output,
    Resource,
)
from delta.manifest.v1_1 import ManifestParserV1v1


class ManifestParserV1v4(ManifestParser):
    @classmethod
    def parse_license(cls, data: dict | None) -> License:
        if data is None:
            return None
        return ManifestParserV1v1.parse_license(data)

    @classmethod
    def parse_resource(cls, data: dict) -> Resource:
        return ManifestParserV1v1.parse_resource(data)

    @classmethod
    def parse_input(cls, data: dict) -> Input:
        return ManifestParserV1v1.parse_input(data)

    @classmethod
    def parse_output(cls, data: dict) -> Output:
        return ManifestParserV1v1.parse_output(data)

    @classmethod
    def parse_model(cls, data: dict) -> Model:
        return ManifestParserV1v1.parse_model(data)

    @classmethod
    def parse_dependency(cls, data: dict) -> Dependency:
        return Dependency(**data)

    @classmethod
    def parse(cls, data: dict) -> Manifest:
        return Manifest(
            name=data["name"],
            owner=data["owner"],
            description=data["description"],
            short_description=data.get("short_description", None),
            license=cls.parse_license(data.get("license", None)),
            resources={
                k: cls.parse_resource({"name": k, **v})
                for k, v in data.get("resources", {}).items()
            },
            inputs={
                k: cls.parse_input({"name": k, **v})
                for k, v in data.get("inputs", {}).items()
            },
            outputs={
                k: cls.parse_output({"name": k, **v})
                for k, v in data.get("outputs", {}).items()
            },
            models={
                k: cls.parse_model(v)
                for k, v in data.get("models", {}).items()
            },
            dependencies={
                k: cls.parse_dependency(v)
                for k, v in data.get("dependencies", {}).items()
            },
        )



import socketio
from rsa import PublicKey, PrivateKey, encrypt, decrypt, pkcs1
from delta.run.config import Settings

_settings: Settings = None


def _get_socketio_adapter_url():
    global _settings
    if _settings is None:
        _settings = Settings()
    return _settings.socketio_adapter_url


def encrypt_with_rsa(value: str, public_key: PublicKey) -> bytes:
    """
    Encrypts a string using the provided public key.
    """
    try:
        return encrypt(value.encode(), public_key)
    except pkcs1.CryptoError as e:
        raise ValueError(f"Encryption failed: {e}")


def decrypt_with_rsa(value: bytes, private_key: PrivateKey) -> str:
    """
    Decrypts a string encrypted with the corresponding private key.
    """
    try:
        decrypted_value = decrypt(value, private_key).decode("utf-8")
        return decrypted_value
    except pkcs1.CryptoError as e:
        raise ValueError(f"Decryption failed: {e}")


class DependencyResolver:
    def __init__(self):
        self.dependencies = {}
        self.visited = set()

    def add_dependency(self, component, dependency):
        self.dependencies.setdefault(component, []).append(dependency)

    def check_dependency_cycles(self, node):
        try:
            self.visited.clear()
            self.check_cycle(node, set())
            return False
        except ValueError:
            return True

    def check_cycle(self, node, visited):

        if node in visited:
            raise ValueError("Dependency cycle detected, Unable to continue.")

        visited.add(node)

        dependencies = self.dependencies.get(node, [])
        for neighbor in dependencies:
            self.check_cycle(neighbor, visited.copy())


class NotifierManager:
    __instance = None

    def __new__(cls):
        if cls.__instance is None:
            cls.__instance = super(NotifierManager, cls).__new__(cls)
            cls.__instance._init()
        return cls.__instance

    def _init(self):
        self._users = {}
        self._notifier = socketio.AsyncRedisManager(
            url=_get_socketio_adapter_url(),
            write_only=True,
        )

    def save_user_sid(self, username, sid):
        self._users[username] = sid

    def delete_user_sid(self, sid):
        for key, value in self._users.copy().items():
            if value == sid:
                del self._users[key]

    async def notify_user(self, username, event, data):
        user = self._users.get(username)
        if user:
            await self._notifier.emit(event, data, room=username)



import logging
import os
from typing import Any

from delta.drive import DeltaDrive
from delta.manifest.manifest import read_manifest
from delta.runners.factory import create
from delta.vcs.handler import GitlabAPI

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("Delta")


class DeltaCore:
    def __init__(self):
        """
        TODO: Need to change IDrive and IRun
        => Need a real instance.
        """
        self._drive = DeltaDrive()
        self._run = None

    def drive_clone(self, url: str, path_to_save: str = "."):
        self._drive.clone(url, path_to_save)

    def drive_init(self, path_to_save: str):
        self._drive.init(path_to_save)

    def drive_status(self):
        self._drive.status()

    def drive_tag(self, **kwargs):
        return self._drive.tag(**kwargs)

    def drive_commit(self, message: str):
        self._drive.commit(message)

    def drive_pull(self, origin_name: str = ""):
        self._drive.pull(origin_name)

    def drive_push(self, **kwargs):
        self._drive.push(**kwargs)

    def drive_branch(self, branch_name: str = None, delete: bool = False):
        """
        Create/Remove branch
        """
        if branch_name is None or branch_name == "":
            return self._drive.branch()
        else:
            self._drive.branch(branch_name=branch_name, delete=delete)

    def drive_checkout(self, branch_name: str):
        self._drive.checkout(branch_name)

    def drive_reset(self):
        self._drive.reset()

    def drive_add_dependency(self):
        self._drive.add_dependency()

    def drive_add_resource(self, **kwargs):
        self._drive.add_resource(**kwargs)
        logger.info("A resource has been added.")

    def drive_delete_resource(self, **kwargs):
        self._drive.delete_resource(**kwargs)

    def drive_get_resources(self):
        return self._drive.get_resources()

    def drive_check(self):
        return self._drive.check()

    def drive_fetch(self, origin_name: str = ""):
        self._drive.fetch(origin_name)

    def drive_sync(self):
        self._drive.sync()

    def run_configure(self, **kwargs):
        if self._run is None:
            logger.warning("Need a run instance to configure the runner.")
            return
        _manifest = read_manifest(
            os.path.join(self._drive.repo_directory, "manifest.json")
        )
        kwargs["resources"] = _manifest["resources"]
        self._run.configure(**kwargs)

    def run_start(self, **kwargs):
        if self._run is None:
            _manifest = read_manifest(
                os.path.join(self._drive.repo_directory, "manifest.json")
            )
            self._run = create(_manifest["models"])
            kwargs["resources"] = _manifest["resources"]
        logger.info("Job runner is starting...")
        self._run.start(**kwargs)
        logger.info("Done.")

    def run_stop(self, **kwargs):
        if self._run is None:
            _manifest = read_manifest(
                os.path.join(self._drive.repo_directory, "manifest.json")
            )
            self._run = create(_manifest["models"])
        logger.info("Stopping runner...")
        self._run.stop(**kwargs)
        logger.info("Done.")

    def run_resume(self, **kwargs):
        if self._run is None:
            logger.warning("Cannot resume, need to start before.")
            return
        _manifest = read_manifest(
            os.path.join(self._drive.repo_directory, "manifest.json")
        )
        kwargs["resources"] = _manifest["resources"]
        self._run.resume(**kwargs)

    def run_monitor(self, **kwargs):
        if self._run is None:
            logger.warning("Need an instance to monitor the runner.")
            return
        _manifest = read_manifest(
            os.path.join(self._drive.repo_directory, "manifest.json")
        )
        kwargs["resources"] = _manifest["resources"]
        self._run.monitor(**kwargs)

    def list(self, url, private_token):
        with GitlabAPI(url, private_token) as glapi:
            glapi.get_projects()

    def drive_build(self, version, registry, no_cache=False):
        self._drive.docker_build(
            version=version, registry=registry, no_cache=no_cache
        )

    def set_verbosity_level(self, verbose_level):
        logger.setLevel(verbose_level)

    def close(self):
        pass

    def __enter__(self) -> "DeltaCore":
        return self

    def __exit__(self, *args: Any, **kwargs) -> None:
        self.close()



import sys
import os
import json
import jsonschema
from delta.manifest.constants import manifest_version
from delta.manifest.constants import manifest_schema_filename

manifest_minimum_sample = "minimum.json"


class ManifestException(Exception):
    """Raised when Error caught in manifest."""
    pass


def _schema_path(version=manifest_version):
    version = 'v' + version.replace('.', "_")
    return os.path.join(os.path.dirname(__file__), version)


def _find_schema_path(version=manifest_version,
                      schema_filename=manifest_schema_filename) -> str:
    f"""
    Build the schema path according to its version.
    The file presence is not checked here.
    :param version: the version of the schema (default :{manifest_version})
    :param schema_filename: the filename retrieved form this package
      (default: {manifest_schema_filename})
    :return: the full path to the schema file
    """
    # retrieve the schema source path
    # Expected schemas are stored into their version folders.
    if version is None:
        version = manifest_version
    if schema_filename is None:
        schema_filename = manifest_schema_filename

    return os.path.join(_schema_path(version), schema_filename)


def _read_json_file(path):
    with open(path) as fp:
        return json.load(fp)


def read_manifest(path='manifest.json'):
    """
    Read the manifest json file to return dict instance of the decoded file.
    :param path: the path to the manifest
    :return: the json dict instance
    """
    # read the given manifest
    return _read_json_file(path)


def write_manifest(path='manifest.json', content=None,
                   version=manifest_version):
    """
    Write the manifest into path destination.
    :param path: the path to the manifest. If the path already exist, file is
       overwritten.
    :param content: the manifest content dict (if provided, this content is
       not checked before written)
    :param version: the version of the manifest to be written (ignored if
       content is provided).
    :return: the generated manifest path
    """
    if not content:
        try:
            content = _read_json_file(
                os.path.join(os.path.dirname(
                    _find_schema_path(version=version)),
                    manifest_minimum_sample))
        except FileNotFoundError as e:
            raise ManifestException(
                f"Minimum manifest content not found for "
                f"version \"{version}\"") from e
    # read the given manifest
    with open(path, "w") as outfile:
        json.dump(content, outfile)

    return path


def check_manifest(data: dict, verbose=False, schema=None, version=None) \
        -> bool:
    f"""
    Silently or verbosely check the manifest. The schema used for validation
     is the default defined into constants.current_version module
     ({manifest_version}).

    :param data: content to validate.
    :param verbose: verbose check if true, no output otherwise.
    :param schema: provide an other json schema to check the manifest.
    :param version: define the schema version for validation (ignored if
       schema is provided).
    :return: True if manifest of valid, false otherwise.
    """
    if schema is None:
        schema = _read_json_file(_find_schema_path(version=version))

    validator = jsonschema.validators.validator_for(schema)(schema)
    if verbose:
        errors = sorted(validator.iter_errors(data), key=lambda e: e.path)
        for error in errors:
            print(f"- {error.message}")
        return not errors
    else:
        return validator.is_valid(data)


if __name__ == '__main__':
    globals()[sys.argv[1]]()



import unittest
import os
from delta.vcs.handler.delta_git_handler import DeltaGitHandler


class TestInit(unittest.TestCase):
    def test_init(self):
        gh = DeltaGitHandler()
        gh.create('test_init')
        self.assertEqual(gh._path, os.path.abspath('test_init'))
        self.assertEqual(gh._initialize_flag, True)
        self.assertIsNotNone(gh._repo)

        self.assertTrue(os.path.exists(os.path.join(
                                            os.path.abspath('test_init'),
                                            'manifest.json')))
        self.assertTrue(os.path.exists(os.path.join(
                                            os.path.abspath('test_init'),
                                            '.delta')))
        self.assertTrue(os.path.exists(os.path.join(
                                            os.path.abspath('test_init'),
                                            'resources')))
        self.assertTrue(os.path.exists(os.path.join(
                                            os.path.abspath('test_init'),
                                            'artifacts')))
        self.assertTrue(os.path.exists(os.path.join(
                                            os.path.abspath('test_init'),
                                            'sources')))
        self.assertTrue(os.path.exists(os.path.join(
                                            os.path.abspath('test_init'),
                                            'models')))



import unittest
from delta.vcs.handler.delta_git_handler import DeltaGitHandler, DeltaException


class TestRemote(unittest.TestCase):
    def test_remote_list(self):
        gh = DeltaGitHandler()
        gh.create('test_remote_list')
        gh.create_remote('origin', 'afakeurlwhichnormallydontwork.com')
        self.assertEqual(gh.list_remotes(), ['origin'])

        gh.create_remote('toto', 'afakeurlwhichnormallydontwork2.com')

        self.assertEqual(gh.list_remotes(), ['origin', 'toto'])

    def test_remote_list_empty(self):
        gh = DeltaGitHandler()
        gh.create('test_remote_list_empty')
        self.assertEqual(gh.list_remotes(), [])

    def test_remote_list_not_initialized(self):
        gh = DeltaGitHandler()
        with self.assertRaises(DeltaException):
            gh.list_remotes()



import datetime as dt
import uuid

import docker.errors

from delta.run.db.orm import RunStatus
from delta.run.job import Job, JobService


class DockerJobService(JobService):
    def __init__(self, registry: str, keep_container: bool = False):
        super().__init__()
        self._registry = registry
        self._cli = docker.from_env()
        self._keep_container = not keep_container
        self._last_check = int(dt.datetime.now(dt.timezone.utc).timestamp())

    async def execute_job(self, job: Job):
        tag = f"{self._registry}/{job.image.tag}"
        volumes = {e.src: {"bind": e.dest, "mode": "ro"} for e in job.inputs}
        if job.outputs:
            path = job.outputs[0].dest
            volumes[path] = {"bind": f"/outputs/{str(job.id)}", "mode": "rw"}
        try:
            outputs = " ".join([e.src for e in job.outputs])
            final_command = job.command
            if outputs:
                final_command += f" && cp {outputs} /outputs/{str(job.id)}"
            cmd = [
                "/bin/bash",
                "-c",
                final_command
            ]
            self._cli.containers.run(
                image=tag,
                command=cmd,
                volumes=volumes,
                detach=True,
                labels={"app": "delta-run", "jobId": str(job.id)},
                remove=self._keep_container
            )
            self._jobs[job.id] = job
        except Exception as ex:
            self._logger.error(
                "Failed to execute job %s: %s", str(job.id), str(ex)
            )
            job.status = RunStatus.ERROR

    def check_jobs(self) -> None:
        now = int(dt.datetime.now(dt.timezone.utc).timestamp())

        for event in self._cli.events(
            since=self._last_check,
            until=now,
            decode=True,
            filters={"label": "app=delta-run"}
        ):
            attributes = event["Actor"]["Attributes"]
            # retrieve associated job
            job_id = uuid.UUID(attributes["jobId"])
            job = self._jobs.get(job_id)
            if job is None:
                self._logger.warning("Unknown job %s", job_id)
            # check event
            if event["status"] == "die":
                rc = int(attributes["exitCode"])
                job.status = RunStatus.SUCCESS if rc == 0 else RunStatus.ERROR
                del self._jobs[job_id]

        # update last check time
        self._last_check = now

    def shutdown(self):
        self._cli.close()



import requests
from typing import Any


class GitlabAPI:

    def __init__(self,
                 url: str,
                 private_token: str = '',
                 authorization_type='Bearer'):
        self._url = url
        self._private_token = private_token
        self._authorization_type = authorization_type

    def _get_requests(self, api_path: str):
        headers = {
            "Content-Type": "application/json",
            'Authorization':    f'{self._authorization_type} '
                                f'{self._private_token}'
        }
        return requests.get(
                    self._url + api_path,
                    headers=headers)

    def _post_requests(self, api_path: str, data):
        headers = {
            'Authorization':    f'{self._authorization_type} '
                                f'{self._private_token}'
        }
        return requests.post(
                        self._url + api_path,
                        headers=headers,
                        json=data)

    def _put_requests(self, api_path: str, data):
        headers = {
            "Content-Type": "application/json",
            'Authorization':    f'{self._authorization_type} '
                                f'{self._private_token}'
        }
        return requests.put(
                        self._url + api_path,
                        headers=headers,
                        json=data
                        )

    def _delete_request(self, api_path):
        headers = {
            'Authorization':    f'{self._authorization_type} '
                                f'{self._private_token}'
        }
        return requests.delete(
            self._url + api_path,
            headers=headers
        )

    def get_topics(self):
        return self._get_requests('/api/v4/topics')

    def get_topic_by_id(self, topic_id):
        return self._get_requests(f'/api/v4/topics/{topic_id}')

    def get_topic_by_name(self,
                          topic_name: str):
        res = self.get_topics()
        if res.status_code == 200:
            topics = res.json()
            for topic in topics:
                if topic["name"] == topic_name:
                    return topic
            return None

    def get_projects(self):
        return self._get_requests('/api/v4/projects')

    def get_namespace_id(self, namespace_full_path):
        response = self._get_requests('/api/v4/groups')
        if response.status_code == 200:
            content = response.json()
            for group in content:
                if (group['full_path'] == namespace_full_path):
                    return group['id']
        return None

    def get_projects_by_namespace_full_path(self, namespace_full_path):
        namespace_id = self.get_namespace_id(namespace_full_path)
        if namespace_id is not None:
            return self._get_requests('/api/v4/groups/'
                                      f'{namespace_id}/projects')
        return None

    def get_readme_of_project(self, project_id: int):
        return self._get_requests(f'/api/v4/projects/{project_id}/'
                                  'repository/files/README%2Emd/raw')

    def get_project_by_topic(self, topic_name):
        return self._get_requests(f'/api/v4/projects/?topic={topic_name}')

    def get_project_by_id(self, project_id: int):
        return self._get_requests(f'/api/v4/projects/{project_id}')

    def get_project_tags(self, project_id):
        return self._get_requests(f'/api/v4/projects/{project_id}'
                                  '/repository/tags')

    def create_topic(self,
                     name: str,
                     title: str,
                     description: str = '',
                     **kwargs):
        avatar = kwargs.get('avatar', None)
        print(f'Avatar {avatar} not used.')
        data = {
            "name": name,
            "title": title,
            "description": description
        }
        return self._post_requests('/api/v4/topics', data=data)

    def update_topic(self,
                     topic_id: int,
                     **kwargs):
        res = self.get_topic_by_id(topic_id)
        if res.status_code == 200:
            topic = res.json()
            topic['name'] = kwargs.get('name', topic['name'])
            topic['title'] = kwargs.get('title', topic['title'])
            topic['description'] = kwargs.get(
                                        'description',
                                        topic['description'])

            return self._put_requests(
                                f'/api/v4/topics/{topic_id}',
                                topic
                        )
        else:
            print(f'Update topic error : {res.status_code}')

    def add_topic_to_project(self, project_id, topic_name):
        res = self.get_project_by_id(project_id)
        if res.status_code == 200:
            project = res.json()
            project_topics = project['topics']
            project_topics.append(topic_name)
            data = {
                "topics": project_topics
            }
            return self._put_requests(f'/api/v4/projects/{project_id}', data)
        else:
            print(f"Add topic error : {res.status_code}")

    def remove_topic_to_project(self, project_id, topic_name):
        res = self.get_project_by_id(project_id)
        if res.status_code == 200:
            project = res.json()
            project['topics'].remove(topic_name)
            data = {
                'topics': project['topics']
            }
            return self._put_requests(f'/api/v4/projects/{project_id}', data)
        else:
            print(f"Remove topic error : {res.status_code}")

    def delete_topic(self, topic_id):
        res = self._delete_request(f'/api/v4/topics/{topic_id}')
        return res

    def download_project(self, project_id, sha):
        res = self._get_requests(f'/api/v4/projects/{project_id}/'
                                 f'repository/archive.zip?sha={sha}')
        return res

    def download_project_url(self, project_id, sha):
        return (f'{self._url}/api/v4/projects/{project_id}/'
                f'repository/archive.zip?sha={sha}')

    def close(self):
        pass

    def __enter__(self) -> "GitlabAPI":
        return self

    def __exit__(self, *args: Any, **kwargs) -> None:
        self.close()



from delta.manifest.parser import (
    Input, InputModel, License, Manifest, ManifestParser, Model, Output,
    OutputModel, Resource
)
from delta.manifest.v1_0 import ManifestParserV1v0


class ManifestParserV1v1(ManifestParser):
    @classmethod
    def parse_license(cls, data: dict) -> License:
        return ManifestParserV1v0.parse_license(data)

    @classmethod
    def parse_resource(cls, data: dict) -> Resource:
        return Resource(**data)

    @classmethod
    def parse_input(cls, data: dict) -> Input:
        return Input(**data)

    @classmethod
    def parse_output(cls, data: dict) -> Output:
        return Output(**data)

    @classmethod
    def parse_model(cls, data: dict) -> Model:
        return Model(
            path=data["path"],
            type=data["type"],
            parameters=data["parameters"],
            inputs={
                k: InputModel(**{"name": k, **v})
                for k, v in data.get("inputs", {}).items()
            },
            outputs={
                k: OutputModel(**{"name": k, **v})
                for k, v in data.get("outputs", {}).items()
            }
        )

    @classmethod
    def parse(cls, data: dict) -> Manifest:
        return Manifest(
            name=data["name"],
            description=data["description"],
            license=cls.parse_license(data["license"]),
            resources={
                k: cls.parse_resource({"name": k, **v})
                for k, v in data.get("resources", {}).items()
            },
            inputs={
                k: cls.parse_input({"name": k, **v})
                for k, v in data.get("inputs", {}).items()
            },
            outputs={
                k: cls.parse_output({"name": k, **v})
                for k, v in data.get("outputs", {}).items()
            },
            models={
                k: cls.parse_model(v)
                for k, v in data.get("models", {}).items()
            },
            dependencies=data.get("dependencies", [])
        )



"""update_run_table

Revision ID: 2e57fc6dd873
Revises: 825d896d998a
Create Date: 2024-03-04 15:03:48.187472

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "2e57fc6dd873"
down_revision: Union[str, None] = "825d896d998a"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "runs", sa.Column("return_code", sa.Integer(), nullable=True)
    )
    op.add_column("runs", sa.Column("message", sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("runs", "message")
    op.drop_column("runs", "return_code")
    # ### end Alembic commands ###



import abc
import asyncio
import dataclasses
import logging
import uuid
from collections import namedtuple
from contextlib import suppress
from delta.run.db.orm import RunStatus
from typing import Callable, List


PathMapper = namedtuple("PathMapper", ["src", "dest"])


@dataclasses.dataclass
class Image:
    tag: str
    registry_url: str = dataclasses.field(default=None)
    registry_username: str = dataclasses.field(default=None)
    registry_password: str = dataclasses.field(default=None)


@dataclasses.dataclass
class Job:
    command: str
    image: Image
    id: uuid.UUID = dataclasses.field(default_factory=uuid.uuid4)
    working_dir: str = dataclasses.field(default=None)
    inputs: List[PathMapper] = dataclasses.field(default_factory=list)
    outputs: List[PathMapper] = dataclasses.field(default_factory=list)
    _status: RunStatus = dataclasses.field(
        init=False, default=RunStatus.CREATED
    )
    _callbacks: List[Callable] = dataclasses.field(
        init=False, default_factory=list
    )

    @property
    def status(self) -> RunStatus:
        return self._status

    @status.setter
    def status(self, new: RunStatus):
        old = self._status
        self._status = new
        for callback in self._callbacks:
            callback(old, new)

    def add_callback(self, callback: Callable):
        self._callbacks.append(callback)

    def __repr__(self):
        return (
            f"Job(command={self.command!r}, image={self.image!r}, "
            f"inputs={self.inputs!r}, outputs={self.outputs!r})"
        )

    def __str__(self):
        return self.__repr__()


class JobChecker:
    def __init__(self, func, time):
        self.func = func
        self.time = time
        self.started = False
        self._task = None

    async def start(self):
        if not self.started:
            self.started = True
            self._task = asyncio.ensure_future(self._check())

    async def stop(self):
        if self.started:
            self.started = False
            self._task.cancel()
        with suppress(asyncio.CancelledError):
            await self._task

    async def _check(self):
        while True:
            await asyncio.sleep(self.time)
            self.func()


class JobService(abc.ABC):
    def __init__(self, check_interval: int = 5):
        self._logger = logging.getLogger("JobService")
        self._jobs: dict[uuid.UUID, Job] = {}
        self.__interval = check_interval
        self.__checker = None

    async def __aenter__(self):
        if self.__checker is None:
            self.__checker = JobChecker(self.check_jobs, self.__interval)
        await self.__checker.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.__checker.stop()

    @abc.abstractmethod
    async def execute_job(self, job: Job):
        raise NotImplementedError

    @abc.abstractmethod
    def check_jobs(self) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def shutdown(self):
        raise NotImplementedError



from abc import ABC, abstractmethod
from typing import TypeVar
from .i_data import IData

T = TypeVar('T')


class IDataIO(ABC):
    def __init__(self, src: IData, dst=None):
        self._src = src
        self._dst = dst

    @abstractmethod
    def _read(self) -> T:
        raise NotImplementedError()

    @abstractmethod
    def _write(self, data: T):
        raise NotImplementedError()
